# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles import aliases to allow backwards compat with backends"""

# from spock.backend.dataclass.args import *
from spock.backend.typed import SavePath
